@@ -1,10 +1,11 @@
+import json 
+
 from volcano.general import variant
 from volcano.general.stddef import Quality
-from volcano.general.xml_reader import XmlReader
+from volcano.general.xml_reader import XmlReader, LoadException
 
 from . import time_tools
 from . import base_dev
-from . import merk_defs
 from . import connected_values
 from .merk_nzif_common import ProtocolException
 
@@ -21,6 +22,8 @@ MEASURES_ENERGY = (
         uicos_14    2.3.14.2, чтение токов, напряжений и углов запросом с кодом 14
 '''
 
+# {'tags': ['Ia', 'Ib', 'Ic'], 'f': 0.001, 'bwri': [2, None, None]},
+
 MERK230_TYPE_DEF = {
     'msr-pqs': [
         {'tags': ['P', 'P1', 'P2', 'P3'], 'f': 0.01, 'bwri': [0, 0, 0], 'pq_sign': 'P'},
@@ -28,9 +31,13 @@ MERK230_TYPE_DEF = {
         {'tags': ['S', 'S1', 'S2', 'S3'], 'f': 0.01, 'bwri': [0, 2, 0], 'pq_sign': None}
     ],
     'msr-uicos': [
-        {'tags': ['Ia', 'Ib', 'Ic'], 'f': 0.001, 'bwri': [2, None, None]},
         {'tags': ['Ua', 'Ub', 'Uc'], 'f': 0.01, 'bwri': [1, None, None]},
         {'tags': ['CosAB', 'CosAC', 'CosBC'], 'f': 0.001, 'bwri': [5, None, None]}
+    ],
+    'fn_08_11':[
+        {'tag': 'Ia', 'f': 0.001, 'bwri': [2, 0, 1]},
+        {'tag': 'Ib', 'f': 0.001, 'bwri': [2, 0, 2]},
+        {'tag': 'Ic', 'f': 0.001, 'bwri': [2, 0, 3]},
     ]
 }
 
@@ -43,12 +50,26 @@ class MerkDevice(base_dev.BaseDevice):
     def __init__(self, my_xml_node, proposed_aux_name, parent_log):
         super().__init__(my_xml_node, proposed_aux_name, parent_log)
 
+        log = self.log()
+        
         p = XmlReader(my_xml_node)
 
         self.addr_ = p.get_int('slaveNb', min_val=0, max_val=0xff)
 
         # device type
-        self.type_def_ = p.get_dic('type', merk_defs.g_merk_type_defs)
+        # self.type_def_ = p.get_dic('type', merk_defs.g_merk_type_defs)
+        # Правильно было бы сделать default='', но в этом случае теряется совместимость со старыми проектами, где не было json файла,
+        # но в xml был прописан тип 'merk230'
+        type_name = p.get_str('type', default='merk230')
+        if type_name == 'merk230':
+            log.warn('Using builtin Merkury template. To use custom, put json file in work dir and specify the name in "type" parameter. The name should NOT be "merk230" - it is reserved for backward compatability')
+            type_def = MERK230_TYPE_DEF
+        else:
+            try:
+                with open(type_name + '.json') as f:
+                    type_def = json.loads(f.read())
+            except Exception as ex:
+                raise LoadException(ex)
 
         # password
         pwd_str = p.get_str('pwd', '000000')
@@ -76,19 +97,28 @@ class MerkDevice(base_dev.BaseDevice):
             for name in (n1, n2, n3, n4):
                 self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
 
-        self.msr_pqs_ = MERK230_TYPE_DEF['msr-pqs']
+        # self.msr_pqs_ = MERK230_TYPE_DEF['msr-pqs']
+        self.msr_pqs_ = type_def.get('msr-pqs', [])
         for x in self.msr_pqs_:
             nX, n1, n2, n3 = x['tags']
             for name in (nX, n1, n2, n3):
                 if name:  # nX can be None
                     self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
 
-        self.msr_uicos_ = MERK230_TYPE_DEF['msr-uicos']
+        # self.msr_uicos_ = MERK230_TYPE_DEF['msr-uicos']
+        self.msr_uicos_ = type_def.get('msr-uicos', [])
         for x in self.msr_uicos_:
             n1, n2, n3 = x['tags']
             for name in (n1, n2, n3):
                 self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
         
+        # self.fn_08_11_ = MERK230_TYPE_DEF.get('fn_08_11', [])
+        self.fn_08_11_ = type_def.get('fn_08_11', [])
+        for x in self.fn_08_11_:
+            name = x['tag']
+            self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), is_obligatory=False)
+            
+        
         self.values_[FREQ_TAG_NAME] = connected_values.ConnectedValue(self.branch(FREQ_TAG_NAME), variant.FloatValue(), is_obligatory=False)
         self.values_[STATUS_TAG_NAME] = connected_values.ConnectedValue(self.branch(STATUS_TAG_NAME), variant.IntValue(), is_obligatory=False)
                 
@@ -182,6 +212,17 @@ class MerkDevice(base_dev.BaseDevice):
                     log.warning('Error reading {}: {}'.format(comment, e))
                     if not ctx.is_open():
                         raise
+            # 08_11
+            for name, factor, bwri_src in self.fn_08_11_:
+                bwri = calc_bwri(bwri_src)
+
+                comment = 'Read {}'.format(name)
+                try:
+                    self.read_fn_08_11(ctx, bwri, factor, comment, name)
+                except (ConnectionError, TimeoutError, ProtocolException) as ex:
+                    log.warning('Error reading {}: {}'.format(comment, ex))
+                    if not ctx.is_open():
+                        raise
 
             comment = 'Read freq'
             try:
@@ -324,7 +365,7 @@ class MerkDevice(base_dev.BaseDevice):
             b3 = d[off + 1]
             b2 = d[off + 2]
 
-            return b3 | (b2 << 8) | (b1 << 16)
+            return (b1 << 16) | (b2 << 8) | b3
 
 
         log = self.log()
@@ -341,8 +382,7 @@ class MerkDevice(base_dev.BaseDevice):
         ph2_phys = ph2_raw * factor
         ph3_phys = ph3_raw * factor
 
-        log.debug('%s: ph1_raw=%s, ph2_raw=%s, ph3_raw=%s, ph1=%s, ph2=%s, ph3=%s',
-                  comment, ph1_raw, ph2_raw, ph3_raw, ph1_phys, ph2_phys, ph3_phys)
+        log.debug('%s: ph1_raw=%s, ph2_raw=%s, ph3_raw=%s, ph1=%s, ph2=%s, ph3=%s', comment, ph1_raw, ph2_raw, ph3_raw, ph1_phys, ph2_phys, ph3_phys)
 
         ph1_name, ph2_name, ph3_name = tpl
 
@@ -350,6 +390,32 @@ class MerkDevice(base_dev.BaseDevice):
         self.values_[ph2_name].set(ph2_phys)
         self.values_[ph3_name].set(ph3_phys)
         
+    def read_fn_08_11(self, ctx, bwri, factor, comment, name):
+        assert 0 <= bwri <= 0xff, bwri
+        assert isinstance(name, str), name
+
+        def __extract(d) -> int:
+            b1 = d[0]
+            b3 = d[1]
+            b2 = d[2]
+
+            return (b1 << 16) | (b2 << 8) | b3
+
+
+        log = self.log()
+
+        # 2.3.14.2
+        req = bytes([self.addr_, 0x08, 0x11, bwri])
+        ctx.send_rcv(req, 3, comment, log)
+
+        val_raw = __extract(ctx.data())
+
+        val_phys = val_raw * factor
+
+        log.debug('%s: raw=%s, ph=%s', comment, val_raw, val_phys)
+
+        self.values_[name].set(val_phys)
+        
     def read_freq(self, ctx, comment):
 
         log = self.log()
