@@ -1,4 +1,5 @@
 import collections
+import datetime
 
 from volcano.general import variant
 from volcano.general.stddef import Quality
@@ -10,6 +11,8 @@ from . import base_dev
 from . import connected_values
 from .merk_nzif_common import ProtocolException, NackException, BadDataException
 from .nzif_defs import get_nzif_defs
+from .num_ops import bcd_to_int, hi_byte, lo_byte
+from .nzif_profiles import NzifProfiles, MeterInfo
 
 NzifScaling = collections.namedtuple('NzifScaling', ('Ki', 'Ku', 'Ca', 'I_A', 'U_V', 'E_kwth', 'PQS_kwt', 'F_Hz', 'Cos'))
 
@@ -31,19 +34,27 @@ def _get_uint32_ml(ba, offset):
 
 RWRI_SIGN_P = 'P'
 RWRI_SIGN_Q = 'Q'
+VAR_NAME_DATETIME = 'date_time'
 
 
 class NzifDevice(base_dev.BaseDevice):
 
     def __init__(self, my_xml_node, proposed_aux_name, parent_log):
         super().__init__(my_xml_node, proposed_aux_name, parent_log)
+        
+        log = self.log()
 
         p = XmlReader(my_xml_node)
 
         self.addr_ = p.get_int('slaveNb', min_val=0, max_val=0xff)
+        
+        self.time_sync_delta_sec_ = p.get_int('timeSyncDeltaSec', min_val=0, max_val=600, default=30)
+        self.last_sync_date_ = None     # последняя синхронизация. Объект Date()
+        if self.time_sync_delta_sec_ == 0:
+            log.info('Time sync disabled. Use timeSyncDeltaSec="XX" argument for Device to enable.')
 
         # device type
-        self.type_def_ = p.get_dic('type', get_nzif_defs())
+        self.type_def_ = p.get_dic('type', get_nzif_defs(log))
 
         # password
         pwd_str = p.get_str('pwd', '000000')
@@ -56,6 +67,16 @@ class NzifDevice(base_dev.BaseDevice):
         # timeouts
         self.timeout_general_ = time_tools.Timeout(p.get_time_secs('readPeriod', 5.0, min_val=0.0, max_val=600.0), expired=True)
 
+        # profiles
+        enable_profiles = p.get_bool('profiles', default=False)
+        if enable_profiles:
+            mi = MeterInfo(Address=self.addr_, PulseToPhysicalUnitsFactor=1.0, MemorySize=0x10000, MaxNbOfAddrToReadAtOnce=16)
+            self.profiles_ = NzifProfiles(mi)
+            log.info('Profiles reading ON')
+        else:
+            self.profiles_ = None
+            log.info('Profiles reading OFF')
+
         # Create values
         self.msr_1ph_ = self.type_def_['msr-1ph']
         self.msr_3ph_ = self.type_def_['msr-3ph']
@@ -83,6 +104,9 @@ class NzifDevice(base_dev.BaseDevice):
             for name in (n1, n2, n3, n4):
                 self.values_[name] = connected_values.ConnectedValue(self.branch(name), variant.FloatValue(), False)
 
+        self.values_[VAR_NAME_DATETIME] = connected_values.ConnectedValue(self.branch(VAR_NAME_DATETIME), variant.StrValue(), is_obligatory=False)
+
+
     def sync(self, services: dict):
         assert isinstance(services, dict), services
         super().sync(services)
@@ -167,6 +191,22 @@ class NzifDevice(base_dev.BaseDevice):
                     log.warning('Error reading {}: {}'.format(comment, e))
                     if not ctx.is_open():
                         raise
+            # timesync
+            try:
+                self.time_sync(ctx)
+            except (ConnectionError, TimeoutError, ProtocolException) as ex:
+                log.warning('Error during time sync: {}'.format(ex))
+                if not ctx.is_open():
+                    raise
+
+            # profiles
+            try:
+                if self.profiles_:
+                    self.profiles_.read_profiles(ctx, 2, log)
+            except (ConnectionError, TimeoutError, ProtocolException) as ex:
+                log.warning('Error during reading profiles: {}'.format(ex))
+                if not ctx.is_open():
+                    raise
 
         except (ConnectionError, TimeoutError, ProtocolException) as e:
             # Here we get errors only from Ph1, Ph3, En reads when ctx is not recoverable.
@@ -276,17 +316,110 @@ class NzifDevice(base_dev.BaseDevice):
         # бита старшего байта указывают положение вектора полной мощности и должны маскироваться
         # для получения модуля значения параметра. В счетчиках СЭТ-4ТМ.02,03М, ПСЧ-3,4ТМ.05М
         # при чтении мощности потерь старшие биты указывают положение вектора полной мощности потерь.
-        hi_byte = data[item * 3 + 0]
-        hi_byte_data = hi_byte & 0x3f
-        hi_byte_p_sgn = hi_byte & 0x80
-        hi_byte_q_sgn = hi_byte & 0x40
+        hi_b = data[item * 3 + 0]
+        hi_b_data = hi_b & 0x3f
+        hi_b_p_sgn = hi_b & 0x80
+        hi_b_q_sgn = hi_b & 0x40
 
-        value = (hi_byte_data << 16) | (data[item * 3 + 1] << 8) | data[item * 3 + 2]
-        sign_p = 1 if hi_byte_p_sgn == 0 else -1
-        sign_q = 1 if hi_byte_q_sgn == 0 else -1
+        value = (hi_b_data << 16) | (data[item * 3 + 1] << 8) | data[item * 3 + 2]
+        sign_p = 1 if hi_b_p_sgn == 0 else -1
+        sign_q = 1 if hi_b_q_sgn == 0 else -1
 
         return value, sign_p, sign_q
 
+    def time_sync(self, ctx):
+        # 2.4.1.1.1
+        log = self.log_
+        comment = 'Read time'
+        con_val = self.values_[VAR_NAME_DATETIME]
+
+        req = bytes([self.addr_, 0x04, 0x00])
+
+        try:
+            ctx.send_rcv(req, 8, comment, log)
+        except NackException:
+            con_val.invalidate(Quality.QUALITY_NA)
+            raise
+            
+        # Замер системного времени надо сделать максимально близко к чтению времени из счетчика
+        sys_time = datetime.datetime.now()
+
+        try:
+            ss = bcd_to_int(ctx.data(0))        # !ValueError
+            mm = bcd_to_int(ctx.data(1))
+            hh = bcd_to_int(ctx.data(2))
+            D = bcd_to_int(ctx.data(4))
+            M = bcd_to_int(ctx.data(5))
+            Y = bcd_to_int(ctx.data(6)) + 2000
+            
+            dtm = datetime.datetime(Y, M, D, hh, mm, ss)        # !ValueError
+        except ValueError as ex:
+            log.warn(ex)
+            con_val.invalidate(Quality.QUALITY_NA)
+            raise ProtocolException(str(ex))
+        
+        con_val.set(str(dtm))
+
+        # compare 
+        delta = sys_time - dtm
+        delta_secs = int(delta.total_seconds())
+
+        log.debug('Datetime: Meter=%s, System=%s, Delta=%s sec', dtm, sys_time, delta_secs)
+
+        if self.time_sync_delta_sec_ == 0 or self.time_sync_delta_sec_ > abs(delta_secs):
+            log.debug('Time sync disabled for device')
+            return
+        
+        log.debug('Time sync needed: Delta is %s sec [Max allowed = %s sec]', abs(delta_secs), self.time_sync_delta_sec_)
+        
+        if self.last_sync_date_ is not None and self.last_sync_date_ == datetime.date.today():
+            log.debug('Time sync already done today')
+            return
+            
+        if delta_secs > 120:
+            secs_to_add = 120
+        elif delta_secs < -120: 
+            secs_to_add = -120
+        else:
+            secs_to_add = delta_secs
+            
+            
+        # Ограничением выполнения команды является требование коррекции времени без перехода
+        # в следующий или предыдущий час. Если, тем не менее, запрашивается коррекция времени
+        # с переходом в следующий или предыдущий час, то счетчик возвращает в байте состояния обмена
+        # код 01h (недопустимая команда или параметр).
+        # Поскольку коррекция идет макс на 120 сек (2 мин), то проверим минуты
+        new_time = dtm + datetime.timedelta(seconds=secs_to_add)
+        if dtm.hour != new_time.hour:
+            log.debug('Time sync now would result in hour change. Postponing')
+            return
+            
+        # Тут могут быть отрицательные значения, но для нас это неважно: int32 для отрицательных
+        # значений будет FF FF FF XY, а i16 FF XY, то есть отбрасывание старших
+        # байтов валидно.
+        req = bytes([self.addr_, 0x03, 0x0b, hi_byte(secs_to_add), lo_byte(secs_to_add)])
+        comment = 'Adjust time'
+            
+        try:
+            ctx.send_rcv(req, None, comment, log)
+        except NackException as ex:
+            # con_val.invalidate(Quality.QUALITY_NA) Это была запись, нет нужды инвалидировать текущее время
+            # Часы уже корректировались, запомним это. Такой вариант бывает, если Ридер запустить повторно, ведь я пока не сохраняю в БД даты последней коррекции
+            if ex.nack_code == 0x04:
+                log.warn('Time sync failed with nack code 4 (already synced today)')
+                return
+            # Как было описано выше, этот код означает, что была попытка перевести часы в другой час. Хотя у меня и есть
+            # защита от этого, но кто знает.
+            elif ex.nack_code == 0x01:
+                log.warn('Time sync failed with nack code 1 (attempt to change an hour). Just repeat later.')
+                return
+            else:
+                raise
+            
+        log.info('Time successfully synced')
+        self.last_sync_date_ = datetime.date.today()
+        
+
     @staticmethod
     def msr_code_rwri(mode, param, phase):
         return (mode << 4) | (param << 2) | phase
